__author__ = 'saulius'

import pandas as pd
from ..containers import Genes

def read_encode_known_genes(known_genes_filename):
    """
    Reads known_genes file.
    The file can be obtained from the
    ``table browser from ENCODE project <http://encodeproject.org/cgi-bin/hgTables?hgsid=320204017&clade=mammal&org=Human&db=hg19&hgta_group=genes&hgta_track=wgEncodeRegTxn&hgta_table=0&hgta_regionType=genome&position=chr21%3A33%2C031%2C597-33%2C041%2C570&hgta_outputType=wigData&hgta_outFileName=>``

    :param known_genes_filename:
    :return:
    :rtype: pd.DataFrame
    """

    known_genes = pd.read_csv(known_genes_filename, sep='\t')

    cols = list(known_genes.columns)
    cols = map(lambda x : x.strip('#'), cols)

    # Rename cols
    new_cols = []
    for c in cols:
        if c == 'chrom':
            c = 'chromosome'
        elif c == 'txStart':
            c = 'start'
        elif c == 'txEnd':
            c = 'end'
        new_cols.append(c)

    known_genes.columns = new_cols
    known_genes = known_genes.set_index('name')

    return Genes(known_genes)

def read_gtf(gtf_location):
    '''
        Reads GTF File
    :param gtf_location:
    :return:
    '''

    def parse_attribute(attribute_str):
        attribute = attribute_str.split(';')
        attribute = map(lambda x: x.strip(), attribute)
        attribute = map(lambda x: x.split(' '), attribute)
        attribute = filter(lambda x : len(x) == 2, attribute)
        attribute = map(lambda x: (x[0], x[1].strip('"')), attribute)

        return dict(attribute)

    data = pd.read_csv(gtf_location, sep='\t', skiprows=5, header=None)
    data.columns = ['chromosome', 'source', 'feature', 'start', 'end', 'score', 'strand', 'frame', 'attribute']

    # Leave only data about genes
    data = data[data.feature == 'gene']

    # Parse attributes
    parsed_attributes = map(parse_attribute, data['attribute'].values)
    parsed_attributes = pd.DataFrame(parsed_attributes, index=data.index)

    # Join everything into one dataframe
    data = data.join(parsed_attributes)

    # Drop the "attribute" and "feature" columns
    data = data[data.columns - ['attribute', 'feature']]
    data = data.set_index('gene_id')

    # Reorder columns so first_columns are in the beginning of DF
    first_columns = ['seqname', 'start', 'end', 'strand', 'gene_name', 'gene_status', 'gene_type', 'score', 'level']
    data = pd.DataFrame(data, columns=first_columns + list(data.columns - first_columns))

    return Genes(data)


cdef extern from "cdtw.h":
    cdef enum MLPY_DTW_DISTANCES:
        MLPY_DTW_DISTANCE_EUCLIDEAN
        MLPY_DTW_DISTANCE_SQEUCLIDEAN
        MLPY_DTW_DISTANCE_COSINE

    ctypedef struct Path:
       int k
       int *px
       int *py

    void fill_cost_matrix_unconstrained(double *x, double *y, int n, int m, int n_dimensions, int squared, double warping_penalty,
                                        double *cost)
    void fill_cost_matrix_with_sakoe_chiba_constraint(double *x, double *y, int n, int m, int n_dimensions, int squared, double warping_penalty,
                                                      double *cost, int sakoe_chiba_band_parameter)
    void fill_cost_matrix_with_slanted_band_constraint(double *x, double *y, int n, int m, int n_dimensions, int squared, double warping_penalty,
                                                       double *cost, int width)
    void fill_cost_matrix_with_itakura_constraint(double *x, double *y, int n, int m, int n_dimensions, int squared,
                                                  double warping_penalty, double *cost)
    void fill_constrained_cost_matrix(double *x, double *y, int n, int m, int n_dimensions, int squared, double *cost, char *constraint_matrix)

    int path(double *cost, int n, int m, int startx, int starty, Path *p)
    void subsequence(double *x, double *y, int n, int m, double *cost)
    int subsequence_path(double *cost, int n, int m, int starty, Path *p)
    



import unittest
from scipy.spatial.distance import pdist
import numpy as np
from dgw.dtw.distance import dtw_std
from dgw.dtw.parallel import parallel_pdist
from itertools import combinations
from numpy.testing import assert_array_equal

__author__ = 'saulius'

class TestParallelPdist(unittest.TestCase):

    def setUp(self):
        np.random.seed(42)
        self.sample_data = np.random.randn(10, 20)
        self.sample_data_three_dim = self.sample_data.reshape(10, 20, 1)

    def test_result_returned_same_as_scipy_spatial_distance_pdist_one_process(self):
        correct_ans = pdist(self.sample_data, dtw_std)
        parallel_ans = parallel_pdist(self.sample_data_three_dim, n_processes=1)
        assert_array_equal(correct_ans, parallel_ans)

    def test_result_returned_same_as_scipy_spatial_distance_pdist_one_process_with_metric(self):
        correct_ans = pdist(self.sample_data, lambda x, y: dtw_std(x, y, metric='euclidean'))
        parallel_ans = parallel_pdist(self.sample_data_three_dim, n_processes=1, metric='euclidean')
        assert_array_equal(correct_ans, parallel_ans)

    def test_result_returned_same_as_scipy_spatial_distance_pdist(self):
        correct_ans = pdist(self.sample_data, dtw_std)
        parallel_ans = parallel_pdist(self.sample_data_three_dim)
        assert_array_equal(correct_ans, parallel_ans)

    def test_correct_multidim_result(self):

        data = np.random.randn(10, 16, 2)

        correct_ans = np.array([dtw_std(x, y) for x, y in combinations(data, 2)])

        parallel_ans = parallel_pdist(data, n_processes=1)
        assert_array_equal(correct_ans, parallel_ans)














from itertools import izip
from logging import debug

import pandas as pd
import numpy as np

from dgw.data.containers import AlignmentsData
from dgw.dtw.scaling import uniform_scaling_to_length, uniform_shrinking_to_length
from dgw.dtw.utilities import _strip_nans, no_nans_len
from distance import dtw_std


def points_mapped_to(points_on_original_sequence, dtw_path, sequence_a=True):
    """
    Returns all indices on the sequence b that are mapped to the index on sequence_a provided.

    :param points s_on_original_sequence: index of a point on original sequence
    :param dtw_path: DTW warping path between `sequence_a` and `sequence_b`
    :param sequence_a: whether the query point is on `sequence_a` or not. (Will swap sequences otherwise)
    :return:
    """
    if sequence_a:
        path_ours, path_theirs = dtw_path
    else:
        path_theirs, path_ours = dtw_path

    indices = set()
    for p in points_on_original_sequence:
        indices.update(np.nonzero(path_ours == p)[0])
    return path_theirs[sorted(indices)]

def dtw_projection(sequence, base_sequence, dtw_function=dtw_std, path=None):
    """
    Projects given sequence onto a base time series using Dynamic Time Warping
    :param sequence: the sequence that will be projected onto base_sequence
    :param base_sequence: base sequence to project onto
    :param dtw_function: DTW function to compute the path if it is set to None
    :param path: the pre-computed DTW warping path between sequence and base sequence.
    :return: new time series of length base containing x projected on it
    """
    base_sequence = np.asarray(base_sequence)
    sequence = np.asarray(sequence)

    if not path:
        distance, cost, path = dtw_function(sequence, base_sequence, dist_only=False)

    path_other, path_base = path

    current_sums = np.zeros(base_sequence.shape)
    current_counts = np.zeros(base_sequence.shape)

    nnl = no_nans_len(base_sequence)

    try:
        filler = [np.nan] * base_sequence.shape[1]
    except IndexError:
        filler = np.nan

    for i in range(nnl, len(base_sequence)):
        current_sums[i] = filler
        current_counts[i] = filler

    for mapped_i, i in zip(path_base, path_other):
        # Go through the path and sum all points that map to the base location i together
        current_sums[mapped_i] += sequence[i]
        current_counts[mapped_i] += 1

    current_average = current_sums / current_counts

    # Append NaNs as needed
    nans_count = len(base_sequence) - len(base_sequence)
    if nans_count:
        current_average = np.concatenate((current_average,
                                          [[np.nan] * base_sequence.shape[-1]] * (nans_count)))

    return current_average

def dtw_projection_multi(alignments, base, *args, **kwargs):
    """
    DTW Projects given alignments onto the given base
    """
    def _projection(sequence):
        return dtw_projection(sequence, base, *args, **kwargs)

    index = None
    if isinstance(alignments, pd.DataFrame):
        index = alignments.index
        sequences = alignments.values

    new_data = {}

    for index in alignments.items:
        item = alignments.ix[index]
        projection = dtw_projection(item, base, *args, **kwargs)
        new_data[index] = pd.DataFrame(projection, index=range(len(base)), columns=item.columns)

    return AlignmentsData(pd.Panel(new_data))

def dtw_path_averaging(sequence_a, sequence_b, weight_a=1, weight_b=1, path=None, shrink=True, dtw_function=dtw_std):
    """
    Averages the path computed between the two DTW sequences.
    Computes the DTW distance in order to do so.

    :param sequence_a: first sequence to be averaged
    :param sequence_b: second sequence to be averaged
    :param weight_a: weight of first sequence
    :param weight_b: weight of the second sequence
    :param path: computed mapped path between the sequences. Will be computed using `dtw_function` if not provided
    :param shrink: if set to treu the data will be shrinked to length of maximum sequence
    :param dtw_function: function that computes DTW path between two sequences, e.g. see `parametrised_dtw_wrapper`
    :return:
    """
    sequence_a = np.asarray(sequence_a, dtype=float)
    sequence_b = np.asarray(sequence_b, dtype=float)

    if path is None:
        distance, cost, path = dtw_function(sequence_a, sequence_b, dist_only=False)

    path_base, path_other = path

    averaged_path = np.array([(sequence_a[i] * weight_a + sequence_b[j] * weight_b) / (weight_a + weight_b) for i, j in zip(path_base, path_other)])

    if shrink:
        averaged_path = uniform_shrinking_to_length(averaged_path, max(len(_strip_nans(sequence_a)),
                                                                       len(_strip_nans(sequence_b))))

    return averaged_path

def sdtw_averaging(sequence_a, sequence_b, weight_a, weight_b, path=None, shrink=True, dtw_function=dtw_std):
    """
    Implements Scaled Dynamic Time Warping Path Averaging as described in [#Niennattrakul:2009ep]

    .. [#Niennattrakul:2009ep] Vit Niennattrakul and Chotirat Ann Ratanamahatana "Shape averaging under Time Warping",
       2009 6th International Conference on Electrical Engineering/Electronics, Computer, Telecommunications and Information Technology (ECTI-CON)

    :param sequence_a: sequence A
    :param sequence_b: sequence B
    :param weight_a: weight of sequence A
    :param weight_b: weight of sequence B
    :param path: computed mapped path between the sequences. Will be computed using `dtw_function` if not provided
    :param shrink: if set to true the data will be shrinked to the length of maximum seq
    :return:
    """
    sequence_a = np.asarray(sequence_a, dtype=float)
    sequence_b = np.asarray(sequence_b, dtype=float)

    if path is None:
        distance, cost, path = dtw_function(sequence_a, sequence_b, dist_only=False)

    path = izip(path[0], path[1])  # Rezip this for easier traversal

    averaged_path = []

    prev = None

    diagonal_coefficient = int((weight_a + weight_b) / 2.0)  # The paper does not explicitly say how to round this
    for a,b in path:

        item = (weight_a * sequence_a[a] + weight_b * sequence_b[b]) / (weight_a + weight_b)
        if prev is None:
            extension_coefficient = diagonal_coefficient
        else:
            if prev[0] == a:  # The path moved from (i,j-1) to (i,j)
                # assert(prev[1] + 1 == b)
                extension_coefficient = weight_a
            elif prev[1] == b:  # The path moved from (i-1,j) to (i,j)
                # assert(prev[0] + 1 == a)
                extension_coefficient = weight_b
            else:  # Path moved diagonally from (i-1,j-1) to (i,j)
                # assert(prev[0] + 1 == a)
                # assert(prev[1] + 1 == b)
                extension_coefficient = diagonal_coefficient

        new_items = [item] * extension_coefficient
        averaged_path.extend(new_items)
        prev = (a, b)

    averaged_path = np.asarray(averaged_path, dtype=float)
    if shrink:
        averaged_path = uniform_shrinking_to_length(averaged_path, max(len(_strip_nans(sequence_a)),
                                                                       len(_strip_nans(sequence_b))))
    return averaged_path



from dgw._mlpy.dtw import dtw_std as mlpy_dtw_std
import numpy as np
from dgw.dtw.scaling import uniform_scaling_to_length
from dgw.dtw.utilities import _strip_nans, no_nans_len, reverse_sequence


def parametrised_dtw_wrapper(*dtw_args, **dtw_kwargs):
    """
    Returns a wrapper around DTW function with args dtw_args and kwargs dtw_kwargs
    :param dtw_args: positional parameters of DTW
    :param dtw_kwargs: keyword parameters of DTW
    :return:
    """

    def f(x, y, dist_only=False):
        return dtw_std(x, y, dist_only=dist_only, *dtw_args, **dtw_kwargs)

    return f

def dtw_std(x, y, metric='sqeuclidean', dist_only=True, constraint=None, k=None, try_reverse=True, normalise=False,
            scale_first=False, *args, **kwargs):
    """
    Wrapper arround MLPY's dtw_std that supports cleaning up of NaNs, and reversing of strings.
    :param x:
    :param y:
    :param metric: dtw metric to use `sqeuclidean`, `euclidean` or `cosine`
    :param dist_only: return distance only
    :param constraint: constraint of dtw (try `None` or `'slanted_band'`
    :param k: parameter k needed for slanted band constraint
    :param try_reverse: Will try reversing one sequence as to get a better distance
    :param normalise: If set to true, distance will be divided from the length of the longer sequence
    :param scale_first: If set to true, the shorte sequence will be scaled to the length of the longer sequence before DTW
    :param kwargs:
    :return:
    """
    def _normalise(ans, max_len):
        if normalise:
            return ans / max_len
        else:
            return ans

    def _scaled_path(path, scaling_path, flip_paths):
        path_x = np.asarray([scaling_path[i] for i in path[0]])
        path_y = path[1]

        if flip_paths:
            path = (path_y, path_x)
        else:
            path = (path_x, path_y)

        return path

    def _reverse_path(path):
        n = path.max()
        path = n - path
        return path


    x = np.asarray(x, dtype=np.float)
    y = np.asarray(y, dtype=np.float)

    x = _strip_nans(x)
    y = _strip_nans(y)

    max_len = max(len(x), len(y))
    if scale_first:
        if len(x) >= len(y):
            x, y = y, x
            flip_paths = True
        else:
            flip_paths = False

        x, scaling_path = uniform_scaling_to_length(x, len(y), output_scaling_path=True)

    regular_ans = mlpy_dtw_std(x, y, metric=metric, dist_only=dist_only, constraint=constraint, k=k, *args, **kwargs)
    if not try_reverse:
        if dist_only:
            return _normalise(regular_ans, max_len)
        else:
            dist, cost, path = regular_ans
            dist = _normalise(dist, max_len)

            if scale_first:
                path = _scaled_path(path, scaling_path, flip_paths)

            return dist, cost, path
    else:
        reverse_ans = mlpy_dtw_std(reverse_sequence(x), y, metric=metric, dist_only=dist_only, constraint=constraint, k=k, *args, **kwargs)
        if dist_only:
            return _normalise(min(regular_ans, reverse_ans), max_len)
        elif reverse_ans[0] >= regular_ans[0]:
            dist, cost, path = regular_ans
            if scale_first:
                path = _scaled_path(path, scaling_path, flip_paths)
            return _normalise(dist, max_len), cost, path
        else:  # dist_only = False and reverse_ans is smaller
            dist, cost, path = reverse_ans
            path_rev = (_reverse_path(path[0]), path[1])

            if scale_first:
                path_rev = _scaled_path(path_rev, scaling_path, flip_paths)

            cost = np.fliplr(cost)
            return _normalise(dist, max_len), cost, path_rev



from distutils.core import setup, Extension
import os


try:
    from Cython.Build import cythonize
except ImportError:
    cython_supported = False
else:
    cython_supported = True

import numpy
np_lib = os.path.dirname(numpy.__file__)
np_inc = [os.path.join(np_lib, 'core/include')]

cmdclass = {}
# Compile packages from mlpy distribution
if cython_supported:
    ext_modules = [Extension("dgw._mlpy.dtw",
                           ["mlpy_src/dtw/cdtw.c",
                            "mlpy_src/dtw/dtw.pyx"],
                           include_dirs=np_inc)]

    ext_modules = cythonize(ext_modules)
else:
    ext_modules = [Extension("dgw._mlpy.dtw",
                            ["mlpy_src/dtw/cdtw.c",
                             "mlpy_src/dtw/dtw.c"],
                            include_dirs=np_inc)]

setup(
    name='dgw',
    version='0.1.0',
    packages=['dgw', 'dgw.cli', 'dgw.cluster', 'dgw.data', 'dgw.data.parsers', 'dgw.data.visualisation', 'dgw._mlpy',
              'dgw.dtw', 'dgw.tests.data.parsers', 'dgw.tests.data', 'dgw.tests.dtw', 'dgw.tests', 'dgw'],
    requires=['numpy (>=1.6.1)', 'scipy (>=0.10.1)', 'pandas (>= 0.10.1)',
              'pysam (>=0.7.4)', 'fastcluster (>=1.1.7)', 'matplotlib (>= 1.2.0)'
             ],
    scripts=['bin/dgw-explorer', 'bin/dgw-extract-gene-regions', 'bin/dgw-overlaps2poi',
             'bin/dgw-prototypes2dot', 'bin/dgw-worker'],
    ext_modules=ext_modules,
    url='http://sauliusl.github.com/dgw/',
    license='GPLv3',
    author='Saulius Lukauskas',
    author_email='luksaulius@gmail.com',
    description='Dynamic Genome Warping',
    cmdclass=cmdclass,
)




from collections import defaultdict
from logging import debug, error
import os
from matplotlib.patches import Rectangle
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.widgets import Button
from dgw.dtw.visualisation import visualise_dtw, visualise_dtw_mappings


class ClusterPreviewer(object):
    _clusters = None
    _main_gs = None
    _current_cluster_id = None

    root_window = None
    _buttons = None

    _ax_heatmap = None
    _ax_projected_heatmap = None
    _enlarged_heatmap_axis = None
    _enlarged_projected_heatmap_axis = None


    def __init__(self, cluster_roots, output_directory, configuration_filename, cut_level, dtw_function, highlight_colours=None):
        self.output_directory = output_directory
        self.cut_level = cut_level
        self.configuration_filename = configuration_filename

        self._clusters = cluster_roots
        self._initialise_grid()

        self._current_cluster_id = 0

        self.dtw_function = dtw_function

        debug('Calculating projections')
        for c in self._clusters:

            c.ensure_projections_are_calculated()
            c.ensure_points_of_interest_are_tracked_down()

        self._buttons = []
        self.highlight_colours = highlight_colours

    def gs_prototype(self):
        return self._main_gs[1, 0]
    def gs_projected_mean(self):
        return self._main_gs[1, 1]
    def gs_projected_heatmap(self):
        return self._main_gs[3, :]
    def gs_heatmap(self):
        return self._main_gs[2, :]

    def ax_heatmap(self):
        if self._ax_heatmap is None:
            self._ax_heatmap = plt.subplot(self.gs_heatmap())
        return self._ax_heatmap

    def ax_projected_heatmap(self):
        if self._ax_projected_heatmap is None:
            self._ax_projected_heatmap = plt.subplot(self.gs_projected_heatmap())
        return self._ax_projected_heatmap

    def _initialise_grid(self):
        self._main_gs = gridspec.GridSpec(5, 2, height_ratios=[1, 40, 20, 20, 4], hspace=0.4)


    def current_cluster(self):
        return self._clusters[self._current_cluster_id]

    def _callback_next(self, event):
        debug('next clicked')
        self._current_cluster_id = (self._current_cluster_id + 1) % len(self._clusters)
        self.draw()

    def _callback_previous(self, event):
        debug('previous clicked')
        self._current_cluster_id = (self._current_cluster_id - 1) % len(self._clusters)
        self.draw()

    def _callback_save(self, event):
        debug('Save clicked')

        directory = os.path.join(self.output_directory, '{0}_{1}'.format(self.configuration_filename, self.cut_level))
        if not os.path.exists(directory):
            os.makedirs(directory)

        bed_directory = os.path.join(directory, 'items')
        if not os.path.exists(bed_directory):
            os.makedirs(bed_directory)

        heatmaps_directory = os.path.join(directory, 'heatmaps')
        if not os.path.exists(heatmaps_directory):
            os.makedirs(heatmaps_directory)

        prototypes_directory = os.path.join(directory, 'prototypes')
        if not os.path.exists(prototypes_directory):
            os.makedirs(prototypes_directory)

        print '> Saving regions to directory {0}'.format(bed_directory)
        for i, c in enumerate(self._clusters):

            if c.regions:
                filename = os.path.join(bed_directory, 'cluster_{0}.bed'.format(i + 1))
                c.save_as_encode_track(filename, track_name='DGWCluster{0}'.format(i + 1),
                                       track_description='DGW Cluster {0}/{1} (cut level: {2})'.format(i + 1,
                                                                                                       len(self._clusters),
                                                                                                       self.cut_level))
            else:
                filename = os.path.join(bed_directory, 'cluster_{0}.txt'.format(i + 1))
                c.save_as_list_of_indices(filename)



        print '> Saving heatmaps to directory {0}'.format(heatmaps_directory)
        for i, c in enumerate(self._clusters):

            filename_reg = os.path.join(heatmaps_directory, 'cluster-{0}.pdf'.format(i + 1))
            f = self._plot_regular_heatmap_on_figure(c)
            f.savefig(filename_reg)
            plt.close(f)


            filename_projected = os.path.join(heatmaps_directory, 'cluster-warped-{0}.pdf'.format(i + 1))
            f = self._plot_projected_heatmap_on_figure(c)
            f.savefig(filename_projected)
            plt.close(f)

        print '> Saving prototypes to directory {0}'.format(prototypes_directory)
        for i, c in enumerate(self._clusters):
            filename_prototype = os.path.join(prototypes_directory, 'cluster-{0}.pdf'.format(i + 1))
            f = self._plot_prototype_on_figure(c)
            f.savefig(filename_prototype)
            plt.close(f)

        print '> Saved'

    def add_button(self, text, callback):
        buttons_left = 0.03
        buttons_bottom = 1 - 0.1
        button_width = 0.14
        button_height = 0.05
        button_spacing = 0.02

        ax_button = plt.axes([buttons_left + (button_width + button_spacing) * len(self._buttons),
                              buttons_bottom, button_width, button_height])
        self._buttons.append(Button(ax_button, text))
        self._buttons[-1].on_clicked(callback)

    def _plot_prototype_on_figure(self, cluster):
        f = plt.figure()
        plt.subplots_adjust(left=0.05, bottom=0.05, top=0.95, right=0.95)
        cluster.prototype.plot(ax=f.gca())

        return f

    def _plot_regular_heatmap_on_figure(self, cluster):
        f = plt.figure(figsize=(11.7, 8.3))
        plt.subplots_adjust(left=0.05, bottom=0.05, top=0.95, right=0.95)
        self._enlarged_heatmap_axis = plt.gca()

        cluster.data.plot_heatmap(horizontal_grid=True,
                                          sort_by=None, highlighted_points=cluster.points_of_interest, highlight_colours=self.highlight_colours)

        f.canvas.mpl_connect('button_press_event', self._onclick_listener)
        return f

    def _plot_projected_heatmap_on_figure(self, cluster):
        f = plt.figure(figsize=(11.7, 8.3))
        plt.subplots_adjust(left=0.05, bottom=0.05, top=0.95, right=0.95)
        self._enlarged_projected_heatmap_axis = plt.gca()

        cluster.projected_data.plot_heatmap(horizontal_grid=True,
                                                    sort_by=None, highlighted_points=cluster.tracked_points_of_interest,
                                                    highlight_colours=self.highlight_colours)

        f.canvas.mpl_connect('button_press_event', self._onclick_listener)

        return f

    def _enlarge_heatmap(self, event):
        self._plot_regular_heatmap_on_figure(self.current_cluster())
        plt.show()

    def _enlarge_dtw_heatmap(self, event):
        self._plot_projected_heatmap_on_figure(self.current_cluster())
        plt.show()

    def _view_poi_histogram(self, event):
        current_cluster = self.current_cluster()
        if not current_cluster.points_of_interest:
            plt.figure();
            plt.figtext(0.1, 0.1, 'Sorry, no points of interest for this cluster. Try running explorer again with -poi poi_dataset.bed')
            plt.show();
            return

        # Pandas tend to crash if subplots is True when only one dimension
        number_poi_collections = len(current_cluster.points_of_interest_histogram.columns)
        axes = current_cluster.points_of_interest_histogram.plot(kind='bar', subplots=(number_poi_collections > 1))
        if number_poi_collections == 1:
            axes = [axes]
        for ax in axes:
            ax.set_ylabel('Number of regions')
        plt.xlabel('Bin')
        plt.suptitle('Original points of interest')
        axes = current_cluster.tracked_points_histogram.plot(kind='bar', subplots=(number_poi_collections > 1))
        if number_poi_collections == 1:
            axes = [axes]
        for ax in axes:
            ax.set_ylabel('Number of regions')

        plt.xlabel('Warped bin')
        plt.suptitle('Warped points of interest')
        plt.show()


    def create_buttons(self):
        self.add_button('Previous', self._callback_previous)
        self.add_button('Next', self._callback_next)

        self.add_button('Save all clusters', self._callback_save)

        self.add_button('Enlarge heatmap', self._enlarge_heatmap)
        self.add_button('Enlarge DTW heatmap', self._enlarge_dtw_heatmap)
        self.add_button('View POI Histogram', self._view_poi_histogram)

    def _plot_item_in_figure(self, index):
        data = self.current_cluster().data.ix[index]
        projected_data = self.current_cluster().projected_data.ix[index]
        prototype = self.current_cluster().prototype

        poi = self.current_cluster().points_of_interest.get(index, {})
        tracked_poi = self.current_cluster().tracked_points_of_interest.get(index, {})

        plt.figure()
        ax1 = plt.subplot(3, 1, 1)
        data.plot(ax=ax1, legend=False)
        plt.figlegend(*ax1.get_legend_handles_labels(), loc='lower center')
        if poi:
            lim_min, lim_max = ax1.get_ylim()
            height = (lim_max - lim_min) / 20
            points_plotted_on = defaultdict(lambda: 0)
            for key, value in poi.iteritems():
                colour = self.highlight_colours[key]
                for point in value:
                    items_on_current_point = points_plotted_on[point]
                    ax1.add_patch(Rectangle((point, lim_min + (height*items_on_current_point)), width=1, height=height, facecolor=colour, edgecolor='k'))
                    points_plotted_on[point] += 1
        plt.title('Original')

        ax2 = plt.subplot(3, 1, 2, sharey=ax1)
        prototype.plot(ax=ax2, legend=False)
        plt.title('Cluster Prototype')

        ax3 = plt.subplot(3, 1, 3, sharey=ax1)
        projected_data.plot(ax=ax3, legend=False)

        if tracked_poi:
            lim_min, lim_max = ax3.get_ylim()
            height = (lim_max - lim_min) / 20
            points_plotted_on = defaultdict(lambda: 0)
            for key, value in tracked_poi.iteritems():
                colour = self.highlight_colours[key]
                for point in value:
                    items_on_current_point = points_plotted_on[point]
                    ax3.add_patch(Rectangle((point, lim_min + (height*items_on_current_point)), width=1, height=height,
                                            facecolor=colour, edgecolor='k'))
                    points_plotted_on[point] += 1

        plt.title('Warped')
        plt.suptitle(index)

        figure_dtw_mappings = visualise_dtw_mappings(data, prototype, dtw_function=self.dtw_function,
                                                     columns=data.columns, sequence_x_label=index,
                                                     sequence_y_label='Cluster Prototype')



    def _onclick_listener(self, event):
        if not self.ax_heatmap().contains(event)[0] and not self.ax_projected_heatmap().contains(event)[0]:
           if (self._enlarged_heatmap_axis is None or not self._enlarged_heatmap_axis.contains(event)[0]) and \
               (self._enlarged_projected_heatmap_axis is None or not self._enlarged_projected_heatmap_axis.contains(event)[0]):
               return

        if event.ydata is None or event.button != 3:
            return

        ylabel = event.inaxes.yaxis.get_major_formatter()(event.ydata)
        try:
            self._plot_item_in_figure(ylabel)
            plt.show()
        except Exception, e:
            error('Something went wrong when plotting {0!r}, got: {1!r}'.format(ylabel, e))
            raise

    def draw(self):

        ax_prototype = plt.subplot(self.gs_prototype())
        plt.cla()
        plt.title('Prototype')
        current_cluster = self.current_cluster()
        current_cluster.prototype.plot(ax=ax_prototype, legend=False)

        plt.figlegend(*ax_prototype.get_legend_handles_labels(), loc='lower center')

        self.title.set_text('Cluster #{0}/{2} ({1} elements)'.format(self._current_cluster_id + 1,
                                                                     current_cluster.n_items,
                                                                     len(self._clusters)))

        plt.subplot(self.gs_heatmap())
        plt.cla()
        self._ax_heatmap = plt.gca()

        shared_axis = current_cluster.data.plot_heatmap(horizontal_grid=True, subplot_spec=self.gs_heatmap(),
                                                        sort_by=None, highlighted_points=current_cluster.points_of_interest,
                                                        highlight_colours=self.highlight_colours)

        # Projections
        projections = current_cluster.projected_data

        ax_projections = plt.subplot(self.gs_projected_mean())
        plt.cla()
        plt.title('Average DTW projection onto prototype')
        projections.mean().plot(ax=ax_projections, legend=False)

        plt.subplot(self.gs_projected_heatmap())
        plt.cla()
        self._ax_projected_heatmap = plt.gca()
        current_cluster.projected_data.plot_heatmap(horizontal_grid=True, subplot_spec=self.gs_projected_heatmap(),
                                                    share_y_axis=shared_axis, sort_by=None,
                                                    highlighted_points=current_cluster.tracked_points_of_interest,
                                                    highlight_colours=self.highlight_colours)

        # Finally issue a draw command for the plot
        plt.draw()


    def show(self):
        # A5 Paper size
        self._figure = plt.figure(num=None, figsize=(12, 10), facecolor='w', edgecolor='k')
        self.title = plt.suptitle("") # Create text object for title
        self.create_buttons()
        self._figure.canvas.mpl_connect('button_press_event', self._onclick_listener)
        self.draw()
        plt.show()

class HierarchicalClusteringViewer(object):
    _hierarchical_clustering_object = None
    _gs_main = None
    _figure = None
    _ax_dendrogram = None
    _line = None
    _cut_xdata = 0
    _sorted_index = None
    _last_clusters = None
    _last_xdata = None

    def __init__(self, hierarchical_clustering_object, output_directory, configuration_file, highlight_colours=None):
        self._hierarchical_clustering_object = hierarchical_clustering_object
        self.output_directory = output_directory
        self.configuration_file = configuration_file
        self.highlight_colours = highlight_colours

    @property
    def clusters(self):
        if self._cut_xdata > 0:
            if self._last_xdata != self._cut_xdata:
                index = self.sorted_index
                assert(index is not None)
                self._last_clusters = self._hierarchical_clustering_object.cut_and_resort(self._cut_xdata, index)
                self._last_xdata = self._cut_xdata
            return self._last_clusters
        else:
            return []

    @property
    def gs_dendrogram(self):
        return self.gs_main[1, 0]

    @property
    def gs_heatmap(self):
        return self.gs_main[1, 1]

    @property
    def ax_dendrogram(self):
        return self._ax_dendrogram

    @property
    def gs_main(self):
        return self._gs_main

    @property
    def hierarchical_clustering_object(self):
        return self._hierarchical_clustering_object

    def set_up(self):
        self._gs_main = gridspec.GridSpec(2, 2, wspace=0, height_ratios=[1, 15])
        self._figure = plt.gcf()
        self._ax_dendrogram = plt.subplot(self.gs_dendrogram, rasterized=True)
        self._figure.canvas.mpl_connect('button_press_event', self._onclick_listener)

        self.draw_buttons()

    @property
    def sorted_index(self):
        return self._sorted_index

    def draw_dendrogram(self):
        ax_dendrogram = self.ax_dendrogram
        plt.subplot(ax_dendrogram, rasterized=True)
        hc = self.hierarchical_clustering_object

        dendrogram_dict = hc.dendrogram(orientation='right', get_leaves=True,  distance_sort=True,
                                        color_threshold=self._cut_xdata, ax=ax_dendrogram)
        leaves = dendrogram_dict['leaves']

        plt.setp(plt.gca().get_xticklabels(), rotation='vertical', fontsize=7)
        plt.gca().get_yaxis().set_visible(False)

        if self._sorted_index is None:
            self._sorted_index = hc.data.items[leaves]



    def _callback_preview(self, event):
        if not self.clusters:
            return

        # Else, if we have clusters
        pw = ClusterPreviewer(self.clusters, self.output_directory, self.configuration_file, self._cut_xdata,
                              highlight_colours=self.highlight_colours, dtw_function=self.hierarchical_clustering_object.dtw_function)
        pw.show()

    def _callback_save(self, event):
        pass

    def draw_buttons(self):
        buttons_left = 0.05
        buttons_bottom = 1 - 0.1

        button_width = 0.1
        button_height = 0.05

        button_spacing = 0.02

        ax_button_preview = plt.axes([buttons_left, buttons_bottom, button_width, button_height])
        self.button_preview = Button(ax_button_preview, 'Preview')
        self.button_preview.on_clicked(self._callback_preview)


    def draw(self):
        self.draw_dendrogram()
        hc = self.hierarchical_clustering_object
        sorted_index = self.sorted_index
        assert(sorted_index is not None)

        DENDROGRAM_SCALE = 10  # scipy.cluster.hierarachy.dendrogram scales all y axis values by tenfold for some reason
        hc.data.plot_heatmap(subplot_spec=self.gs_heatmap, no_y_axis=True, sort_by=sorted_index, share_y_axis=self.ax_dendrogram,
                             scale_y_axis=DENDROGRAM_SCALE, highlighted_points=hc.data.points_of_interest, rasterized=True,
                             highlight_colours=self.highlight_colours)

    def show(self):
        # A5 Paper size
        plt.figure(num=None, figsize=(12, 10), facecolor='w', edgecolor='k')
        self.set_up()
        self.draw()
        plt.show()

    def cut_line(self, xdata):

        self._cut_xdata = xdata
        self._colors_list = None
        self.draw_dendrogram()
        if self._line is not None:
            self._line.set_xdata(xdata)
        else:
            self._line = self._ax_dendrogram.axvline(linestyle='--', x=xdata, color='k')

        self._figure.canvas.draw()


    def _onclick_listener(self, event):
        # Allow only in axis, only the left button and only regions with value
        if event.inaxes != self.ax_dendrogram or event.button != 3 or event.xdata is None:
            return

        self.cut_line(event.xdata)




__author__ = 'saulius'
import argparse
import os

class StoreUniqueFilenameAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        seen = set()
        for value in values:
            if value in seen:
                raise argparse.ArgumentError(self, 'Dataset {0!r} is twice in the list'.format(value))
            if not os.path.isfile(value):
                raise argparse.ArgumentError(self, 'File not found {0!r}'.format(value))

            seen.add(value)

        setattr(namespace, self.dest, values)

class StoreFilenameAction(argparse.Action):

    def __call__(self, parser, namespace, value, option_string=None):
        if not os.path.isfile(value):
            raise argparse.ArgumentError(self, 'File not found {0!r}'.format(value))

        setattr(namespace, self.dest, value)

